#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.math.vectors module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest
from pytest import approx

from skhep.math.vectors import *
from skhep.utils.exceptions import *
from math import pi, sqrt


# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------
def test_vectors_constructors():
    with pytest.raises(TypeError):
        Vector3D.fromiterable(1.0)
    with pytest.raises(ValueError):
        Vector3D.fromiterable([1.0])
    with pytest.raises(ValueError):
        Vector3D.fromiterable(["str1", "str2", "str3"])
    #
    v1 = Vector3D()
    assert str(v1) == str((0.0, 0.0, 0.0))
    assert repr(v1) == "<Vector3D (x=0.0,y=0.0,z=0.0)>"
    assert str(v1) == str(Vector3D.origin())
    v2 = Vector3D(1.0, 1.0, 1.0)
    assert str(v2) == str((1.0, 1.0, 1.0))
    v3 = Vector3D.fromvector(v1)
    assert str(v3) == str((0.0, 0.0, 0.0))
    v4 = Vector3D.fromiterable([1.0, 1.0, 1.0])
    assert str(v4) == str((1.0, 1.0, 1.0))
    v5 = Vector3D.fromcylindricalcoords(1.0, 0.0, 1.0)
    assert v5 == Vector3D(1.0, 0.0, 1.0)
    assert v5.rho == 1.0
    v6 = Vector3D.fromcylindricalcoords(0.5, pi / 2, 0.0)
    assert v6 == Vector3D(0.0, 0.5, 0.0)
    v7 = Vector3D.fromsphericalcoords(1.0, 0.0, 0.0)
    assert v7 == Vector3D(0.0, 0.0, 1.0)
    assert v7.r == 1.0
    assert v7.theta() == 0.0
    v8 = Vector3D.fromsphericalcoords(1.0, 0.0, pi / 2)
    assert v8 == Vector3D(0.0, 0.0, 1.0)
    v9 = Vector3D.fromsphericalcoords(2.0, pi / 2, pi / 4)
    assert v9 == Vector3D(sqrt(2), sqrt(2), 0.0)
    v10 = Vector3D.frompoint(2.0, 2.0, 2.0)
    assert str(v10) == str((2.0, 2.0, 2.0))
    #
    with pytest.raises(TypeError):
        LorentzVector.fromiterable(1)
    with pytest.raises(ValueError):
        LorentzVector.fromiterable([1])
    with pytest.raises(ValueError):
        LorentzVector.fromiterable(["str1", "str2", "str3", "str4"])
    #
    lv1 = LorentzVector()
    assert str(lv1) == str((0.0, 0.0, 0.0, 0.0))
    assert repr(lv1) == "<LorentzVector (x=0.0,y=0.0,z=0.0,t=0.0)>"
    lv2 = LorentzVector(1.0, 1.0, 1.0, 1.0)
    assert str(lv2) == str((1.0, 1.0, 1.0, 1.0))
    lv3 = LorentzVector.from4vector(lv1)
    assert str(lv3) == str((0.0, 0.0, 0.0, 0.0))
    lv4 = LorentzVector.from3vector(v2, 2.0)
    assert str(lv4) == str((1.0, 1.0, 1.0, 2.0))
    lv5 = LorentzVector.fromiterable([1.0, 1.0, 1.0, 0.0])
    assert str(lv5) == str((1.0, 1.0, 1.0, 0.0))


def test_containers_properties():
    with pytest.raises(IndexError):
        Vector3D.__setitem__(Vector3D(), 3, 1.0)
    with pytest.raises(IndexError):
        Vector3D.__getitem__(Vector3D(), 3)
    #
    v1 = Vector3D()
    v1.x = 5.0
    v1.y = -5.0
    v1.z = 10
    assert v1 == Vector3D(5.0, -5.0, 10.0)
    v1.set(-5.0, 5.0, 10.0)
    assert v1 == Vector3D(-5.0, 5.0, 10.0)
    v1[0] = 1.0
    assert v1 == Vector3D(1.0, 5.0, 10.0)
    v1[1] = 1.0
    assert v1 == Vector3D(1.0, 1.0, 10.0)
    v1[2] = 1.0
    assert v1 == Vector3D(1.0, 1.0, 1.0)
    assert v1[0] == 1.0
    assert v1[1] == 1.0
    assert v1[2] == 1.0
    assert len(v1) == 3.0
    assert v1.tolist() == [1.0, 1.0, 1.0]
    assert list(v1) == [1.0, 1.0, 1.0]
    assert [v for v in v1] == [1.0, 1.0, 1.0]
    #
    with pytest.raises(IndexError):
        LorentzVector.__setitem__(LorentzVector(), 4, 1.0)
    with pytest.raises(IndexError):
        LorentzVector.__getitem__(LorentzVector(), 4)
    #
    lv1 = LorentzVector()
    lv1.x = 5.0
    lv1.y = -5.0
    lv1.z = 10
    lv1.t = 2.0
    assert lv1 == LorentzVector(5.0, -5.0, 10.0, 2.0)
    lv1.px = 5.0
    lv1.py = 5.0
    lv1.pz = -10
    lv1.e = 2.0
    assert lv1 == LorentzVector(5.0, 5.0, -10.0, 2.0)
    lv1.set(-5.0, 5.0, 10.0, -2.0)
    assert lv1 == LorentzVector(-5.0, 5.0, 10.0, -2.0)
    lv1[0] = 1.0
    assert lv1 == LorentzVector(1.0, 5.0, 10.0, -2.0)
    lv1[1] = 1.0
    assert lv1 == LorentzVector(1.0, 1.0, 10.0, -2.0)
    lv1[2] = 1.0
    assert lv1 == LorentzVector(1.0, 1.0, 1.0, -2.0)
    lv1[3] = 1.0
    assert lv1 == LorentzVector(1.0, 1.0, 1.0, 1.0)
    assert lv1[0] == 1.0
    assert lv1[1] == 1.0
    assert lv1[2] == 1.0
    assert lv1[3] == 1.0
    assert len(lv1) == 4.0
    assert lv1.tolist() == [1.0, 1.0, 1.0, 1.0]
    assert list(lv1) == [1.0, 1.0, 1.0, 1.0]
    assert [v for v in lv1] == [1.0, 1.0, 1.0, 1.0]


def test_vectors_operators():
    with pytest.raises(InvalidOperationError):
        Vector3D.__iadd__(Vector3D(), "a")
    with pytest.raises(InvalidOperationError):
        Vector3D.__isub__(Vector3D(), "a")
    with pytest.raises(InvalidOperationError):
        Vector3D.__imul__(Vector3D(), Vector3D())
    with pytest.raises(ZeroDivisionError):
        Vector3D.__div__(Vector3D(), 0.0)
    with pytest.raises(InvalidOperationError):
        Vector3D.__idiv__(Vector3D(), Vector3D())
    with pytest.raises(ZeroDivisionError):
        Vector3D.__idiv__(Vector3D(), 0.0)
    #
    v1, v2 = Vector3D(0.0, 0.0, 0.0), Vector3D(1.0, 1.0, 1.0)
    v3, v4 = Vector3D(2.0, 2.0, 2.0), Vector3D(3.0, 3.0, 3.0)
    v5 = Vector3D(1.0, 2.0, 3.0)
    assert v1 == 0.0
    assert v1 + v2 == Vector3D(1.0, 1.0, 1.0)
    assert v1 - v2 == Vector3D(-1.0, -1.0, -1.0)
    assert v1 - v2 == -1.0 * Vector3D(1.0, 1.0, 1.0)
    assert v1 * 2.0 == Vector3D(0.0, 0.0, 0.0)
    assert v2 * 2.0 == Vector3D(2.0, 2.0, 2.0)
    assert 2.0 * v1 == Vector3D(0.0, 0.0, 0.0)
    assert 2.0 * v2 == Vector3D(2.0, 2.0, 2.0)
    assert v1 * v2 == 0.0
    assert v2 * v1 == 0.0
    assert v3 * v4 == v3.dot(v4), "Check commutativity"
    assert v3 * v4 == v4.dot(v3), "Check commutativity"
    assert v4 * v3 == v3.dot(v4), "Check commutativity"
    assert v4 * v3 == v4.dot(v3), "Check commutativity"
    assert v3 * v4 == 18.0
    assert v3 / 2.0 == v2
    assert v4 / 3.0 == v2
    v1 *= 2.0
    v2 *= 2.0
    assert v1 == Vector3D(0.0, 0.0, 0.0)
    assert v2 == Vector3D(2.0, 2.0, 2.0)
    v1 /= 2.0
    v2 /= 2.0
    assert v1 == Vector3D(0.0, 0.0, 0.0)
    assert v2 == Vector3D(1.0, 1.0, 1.0)
    assert (v1 + v2) * v3 == 6.0, "Check operations combination"
    assert (v2 - v1) * v3 == 6.0, "Check operations combination"
    assert v3 * (v1 - v2) == -6.0, "Check operations combination"
    assert 18.0 / (v3 * v4) == 1.0, "Check operations combination"
    assert v4 / (v3 * v2) == Vector3D(0.5, 0.5, 0.5)
    assert v2.cross(v2) == Vector3D(0.0, 0.0, 0.0)
    assert v2.cross(v5) == Vector3D(1.0, -2.0, 1.0)
    assert v5.cross(v2) == -1 * v2.cross(v5)
    assert v2 == v2
    assert v2 != v1
    assert not v2 != v2
    assert not v2 == v1
    assert not v1 == "a"
    assert not v1 == 1
    #
    with pytest.raises(InvalidOperationError):
        LorentzVector.__iadd__(LorentzVector(), "a")
    with pytest.raises(InvalidOperationError):
        LorentzVector.__isub__(LorentzVector(), "a")
    with pytest.raises(InvalidOperationError):
        LorentzVector.__imul__(LorentzVector(), LorentzVector())
    with pytest.raises(ZeroDivisionError):
        LorentzVector.__div__(LorentzVector(), 0.0)
    with pytest.raises(InvalidOperationError):
        LorentzVector.__idiv__(LorentzVector(), LorentzVector())
    with pytest.raises(ZeroDivisionError):
        LorentzVector.__idiv__(LorentzVector(), 0.0)
    #
    lv1, lv2 = LorentzVector(0.0, 0.0, 0.0, 0.0), LorentzVector(1.0, 1.0, 1.0, 0.0)
    lv3, lv4 = LorentzVector(2.0, 2.0, 2.0, 1.0), LorentzVector(3.0, 3.0, 3.0, 1.0)
    lv5 = LorentzVector(1.0, 1.0, 1.0, 6.0)
    assert lv1 == 0.0
    assert lv1 + lv2 == LorentzVector(1.0, 1.0, 1.0, 0.0)
    assert lv1 - lv2 == LorentzVector(-1.0, -1.0, -1.0, 0.0)
    assert lv1 * 2.0 == LorentzVector(0.0, 0.0, 0.0, 0.0)
    assert lv2 * 2.0 == LorentzVector(2.0, 2.0, 2.0, 0.0)
    assert 2.0 * lv1 == LorentzVector(0.0, 0.0, 0.0, 0.0)
    assert 2.0 * lv2 == LorentzVector(2.0, 2.0, 2.0, 0.0)
    assert lv2 * lv1 == 0.0
    assert lv1 * lv2 == 0.0
    assert lv3 * lv4 == lv3.dot(lv4)
    assert lv3 * lv4 == lv4.dot(lv3)
    assert lv4 * lv3 == lv3.dot(lv4)
    assert lv4 * lv3 == lv4.dot(lv3)
    assert lv3 * lv4 == -17.0
    assert lv3 / 2.0 == LorentzVector(1.0, 1.0, 1.0, 0.5)
    assert lv4 / 3.0 == LorentzVector(1.0, 1.0, 1.0, 1.0 / 3)
    assert lv3 * lv5 == 0.0
    lv1 *= 2.0
    lv2 *= 2.0
    assert lv1 == LorentzVector(0.0, 0.0, 0.0, 0.0)
    assert lv2 == LorentzVector(2.0, 2.0, 2.0, 0.0)
    lv1 /= 2.0
    lv2 /= 2.0
    assert lv1 == LorentzVector(0.0, 0.0, 0.0, 0.0)
    assert lv2 == LorentzVector(1.0, 1.0, 1.0, 0.0)
    assert lv2 == lv2
    assert lv2 != lv1
    assert not lv2 != lv2
    assert not lv2 == lv1
    assert not lv1 == "a"
    assert not lv1 == 1


def test_vectors_rotations():
    with pytest.raises(TypeError):
        Vector3D.rotate(Vector3D(), pi, 1)
    with pytest.raises(TypeError):
        Vector3D.rotate(Vector3D(), pi, [1, 2])
    with pytest.raises(TypeError):
        Vector3D.rotate(Vector3D(), pi, 1, 2, 3, 4)
    with pytest.raises(ValueError):
        Vector3D.rotate(Vector3D(), pi, 0, 1, "a")
    with pytest.raises(ValueError):
        Vector3D.rotate(Vector3D(), pi, ["a", "b", 3])
    #
    v1 = Vector3D.fromcylindricalcoords(1.0, 0.0, 0.0)
    assert v1.phi() == 0.0
    assert v1.rotatez(pi / 2).phi() == pi / 2
    assert v1.rotatez(pi / 2) == Vector3D(0.0, 1.0, 0.0)
    assert v1.rotatey(pi).phi() == pi
    assert v1.rotatey(-pi).phi() == pi
    assert v1.rotatey(pi) == Vector3D(-1.0, 0.0, 0.0)
    assert v1.rotatey(-pi) == Vector3D(-1.0, 0.0, 0.0)
    assert v1.rotatex(pi).phi() == 0.0
    assert v1.rotatex(pi) == Vector3D(1.0, 0.0, 0.0)
    v2 = Vector3D.fromsphericalcoords(1.0, pi / 2, pi / 2)
    assert v2.phi() == pi / 2
    assert v2.theta() == pi / 2
    assert v2.rotatex(pi).phi() == -pi / 2
    assert v2.rotatex(pi).theta() == pi / 2
    assert v2.rotatex(pi) == Vector3D(0.0, -1.0, 0.0)
    v3 = Vector3D.fromsphericalcoords(1.0, pi / 4, pi / 4)
    angle = v2.angle(v3)
    axis = v2.cross(v3)
    assert v2.rotate(angle, axis) == v3
    assert v2.rotate(-angle, -1.0 * axis) == v3
    #
    with pytest.raises(TypeError):
        LorentzVector.rotate(LorentzVector(), pi, 1)
    with pytest.raises(TypeError):
        LorentzVector.rotate(LorentzVector(), pi, [1, 2])
    with pytest.raises(TypeError):
        LorentzVector.rotate(LorentzVector(), pi, 1, 2, 3, 4)
    with pytest.raises(ValueError):
        LorentzVector.rotate(LorentzVector(), pi, 0, 1, "a")
    with pytest.raises(ValueError):
        LorentzVector.rotate(LorentzVector(), pi, ["a", "b", 3])
    #
    lv1 = LorentzVector.from3vector(v1, 1.0)
    assert lv1.phi() == 0.0
    assert lv1.rotatez(pi / 2).phi() == pi / 2
    assert lv1.rotatez(pi / 2) == LorentzVector(0.0, 1.0, 0.0, 1.0)
    assert lv1.rotatey(pi).phi() == pi
    assert lv1.rotatey(-pi).phi() == pi
    assert lv1.rotatey(pi) == LorentzVector(-1.0, 0.0, 0.0, 1.0)
    assert lv1.rotatey(-pi) == LorentzVector(-1.0, 0.0, 0.0, 1.0)
    assert lv1.rotatex(pi).phi() == 0.0
    assert lv1.rotatex(pi) == LorentzVector(1.0, 0.0, 0.0, 1.0)
    lv2 = LorentzVector.from3vector(v2, 2.0)
    assert lv2.phi() == pi / 2
    assert lv2.theta() == pi / 2
    assert lv2.rotatex(pi).phi() == -pi / 2
    assert lv2.rotatex(pi).theta() == pi / 2
    assert lv2.rotatex(pi) == LorentzVector(0.0, -1.0, 0.0, 2.0)
    lv3 = LorentzVector.from3vector(v3, 2.0)
    assert lv2.rotate(angle, axis) == lv3
    assert lv2.rotate(-angle, -1.0 * axis) == lv3


def test_3Dvectors_properties():
    v0 = Vector3D()
    v1, v2 = Vector3D(1.0, 1.0, 1.0), Vector3D(2.0, 2.0, 2.0)
    v3, v4 = Vector3D(-1.0, -1.0, -1.0), Vector3D(-2.0, -2.0, -2.0)
    v5, v6 = Vector3D(1.0, 1.0, 0.0), Vector3D(0.0, 0.0, 2.0)
    assert not v0
    assert v1
    assert v1.mag2 == 3.0
    assert v1.mag == sqrt(3.0)
    assert v2.mag2 == 12.0
    assert v2.mag == sqrt(12.0)
    assert v1.unit().mag == 1.0
    v7 = v1.unit()
    assert v1.unit() == v7.unit()
    assert v1.isparallel(v2) == True
    assert v2.isparallel(v1) == True
    assert v1.isantiparallel(v2) == False
    assert v2.isantiparallel(v1) == False
    assert v1.isantiparallel(v3) == True
    assert v2.isantiparallel(v4) == True
    assert v1.isparallel(v3) == False
    assert v2.isparallel(v4) == False
    assert v1.isopposite(v3) == True
    assert v1.isopposite(v4) == False
    assert v2.isopposite(v4) == True
    assert v2.isopposite(v3) == False
    assert v1.isperpendicular(v2) == False
    assert v2.isperpendicular(v1) == False
    assert v5.isperpendicular(v6) == True
    assert v6.isperpendicular(v5) == True


def test_lorentz_vectors_properties():
    with pytest.raises(TypeError):
        LorentzVector.boost(LorentzVector(), 1)
    with pytest.raises(TypeError):
        LorentzVector.boost(LorentzVector(), [1, 2])
    with pytest.raises(TypeError):
        LorentzVector.boost(LorentzVector(), 1, 2, 3, 4)
    with pytest.raises(ValueError):
        LorentzVector.boost(LorentzVector(), 0, 1, "a")
    with pytest.raises(ValueError):
        LorentzVector.boost(LorentzVector(), ["a", "b", 3])
    #
    lv0 = LorentzVector()
    lv1, lv2 = LorentzVector(1.0, 1.0, 1.0, 1.0), LorentzVector(1.0, 1.0, 1.0, 2.0)
    assert lv1.boostvector == Vector3D(1.0, 1.0, 1.0)
    assert lv1.beta == sqrt(3.0)
    assert lv2.boostvector == Vector3D(0.5, 0.5, 0.5)
    lv3 = LorentzVector(0.0, 0.0, 1.0, 0.0)
    beta = 0.05
    gamma = 1 / sqrt(1 - beta**2)
    lv4 = lv3.boost(0, 0, beta)
    assert lv4.z == lv3.z * gamma, "Check length contraction"
    assert lv4.x == lv3.x
    assert lv4.y == lv3.x
    assert lv4.t == gamma * (lv3.t - beta * lv3.z)
    assert lv4 == LorentzVector(
        lv3.x, lv3.y, lv3.z * gamma, gamma * (lv3.t - beta * lv3.z)
    )
    lv5 = LorentzVector(0.0, 0.0, 0.0, 1.0)
    assert lv5.beta == 0.0
    assert lv5.gamma == 1.0
    assert lv5.boost(0, 0, 0) == lv5
    lv6 = lv5.boost([0, beta, 0])
    assert lv6.x == lv5.x
    assert lv6.z == lv5.z
    assert lv6.t == lv5.t * gamma, "Check time dilation"
    assert lv6.y == gamma * (lv5.y - beta * lv5.t)
    assert lv6 == LorentzVector(
        lv5.x, gamma * (lv5.y - beta * lv5.t), lv5.z, lv5.t * gamma
    )
    assert lv1.isspacelike() == True
    assert lv1.istimelike() == False
    assert lv1.islightlike() == False
    assert lv2.isspacelike() == False
    assert lv2.istimelike() == True
    assert lv2.islightlike() == False
    lv7 = LorentzVector(1.0, 1.0, 1.0, sqrt(3.0))
    assert lv7.isspacelike() == False
    assert lv7.istimelike() == False
    assert lv7.islightlike() == True
    #
    p1 = LorentzVector()
    p1.setpxpypzm(5.0, 5.0, 10.0, 5)
    assert p1.px == 5.0
    assert p1.py == 5.0
    assert p1.pz == 10.0
    assert p1.m == 5.0
    assert p1.mass == 5.0
    assert p1.m2 == 25.0
    assert p1.mass2 == 25.0
    assert p1.pt == sqrt(p1.px**2 + p1.py**2)
    assert p1.p == sqrt(p1.px**2 + p1.py**2 + p1.pz**2)
    assert p1.p == sqrt(p1.pt**2 + p1.pz**2)
    assert p1.e == sqrt(p1.m**2 + p1.p**2)
    assert p1.beta == p1.p / p1.e
    p2 = p1.boost(p1.px / p1.e, p1.y / p1.e, p1.z / p1.e)
    assert p2 == p1.boost(p1.boostvector)
    assert p2.p == 0.0
    assert p2.m == approx(5.0)
    p3 = LorentzVector()
    p3.setpxpypze(5.0, 5.0, 10.0, 20)
    assert p3.px == 5.0
    assert p3.py == 5.0
    assert p3.pz == 10.0
    assert p3.e == 20.0
    assert p3.m == sqrt(p3.e**2 - p3.p**2)
    assert p3.beta == p3.p / p3.e
    p4 = LorentzVector()
    p4.setptetaphim(10.0, 3.5, pi / 3, 5.0)
    assert p4.pt == approx(10.0)
    assert p4.eta == approx(3.5)
    assert p4.phi() == approx(pi / 3)
    assert p4.m == approx(5.0)
    p5 = LorentzVector()
    p5.setptetaphie(10.0, 3.9, -2 * (pi / 3), 20.0)
    assert p5.pt == approx(10.0)
    assert p5.eta == approx(3.9)
    assert p5.pseudorapidity == approx(3.9)
    assert p5.phi() == approx(-2 * (pi / 3))
    assert p5.e == approx(20.0)
    assert p5.deltaeta(p4) == approx(0.4)
    assert p4.deltaeta(p5) == approx(-0.4)
    assert p5.deltaphi(p4) == approx(-pi)
    assert p4.deltaphi(p5) == approx(pi)
    assert p5.deltar(p4) == approx(sqrt(0.4**2 + pi**2))
    assert p4.deltar(p5) == approx(sqrt(0.4**2 + pi**2))
    p6 = LorentzVector()
    p6.setptetaphie(10.0, 3.9, -pi, 20.0)
    assert p6.deltaphi(p4) == approx(2 * (pi / 3))
    assert p4.deltaphi(p6) == approx(-2 * (pi / 3))
    p7 = LorentzVector()
    p7.setpxpypzm(5.0, 5.0, 5.0, 0.0)
    assert p7.beta == 1.0
    assert p7.gamma == 10e10, "Gamma of the photons is +inf"
    assert p7.p == p7.e, "Momentum = Energy for photons"
    assert p7.pseudorapidity == p7.rapidity
    assert p7.pt == p7.et, "Transverse Momentum = Transverse Energy for photons"
    assert p7.pt == approx(p7.mt), "Transverse Momentum = Transverse Mass for photons"
    assert p7.perp2 == approx(p7.mt2)
    assert p7.islightlike() == True
    p8 = p1 + p7
    p9, p10 = p1.boost(p8.boostvector), p7.boost(p8.boostvector)
    assert p9.p == approx(p10.p), "Check boost to the C.O.M frame"
    p3_9, p3_10 = p9.vector, p10.vector
    assert p3_9.isopposite(p3_10)
    p8 = LorentzVector()
    p8.setptetaphie(10.0, 2e2, -2 * (pi / 3), 20.0)  # sinh(eta) diverge quickly
    assert p8.theta() == 0.0
    assert p8.eta == 10e10



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Module for mathematical functions and utilities.

This module contains in particular:

  * Vector classes.
  * Geometry classes.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from .kinematics import *
from .vectors import *
from .geometry import *



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Mathematical functions relevant to kinematics
=============================================
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from hepunits.units import MeV, ns
from hepunits.constants import hbar

from .numeric import isequal


def Kallen_function(x, y, z):
    """
    The Kallen function, aka triangle or lambda function,
    named after physicist Anders Olof Gunnar Kallen [Kallen]_:

    .. math::

      \\begin{eqnarray}
      \\lambda(x,y,z) &=& x^2 + y^2 + z^2 - 2 x y - 2 y z - 2 z x \\\\
                      &=& (x-y-z)^2 - 4 y z \\\\
                      &=& [ x - (\\sqrt{y}+\\sqrt{z})^2 ] [ x - (\\sqrt{y}-\\sqrt{z})^2 ] \\,\\,\\,\\mathrm{if} \\,\\,\\,y, z > 0
      \\end{eqnarray}

    Example
    -------
    Calculate in the rest frame of a particle of mass M decaying to 2 particles labeled 1 and 2,
    :math:`P (M) \\to p1 (m1) + p2 (m2)`, the momenta of 1 and 2 given by :math:`p = |\\mathbf{p1}| = |\\mathbf{p2}|`:

        >>> from skhep.math  import Kallen_function
        >>> from hepunits.units import MeV, GeV
        >>> from math import sqrt
        >>> M = 5.279 * GeV; m1 = 493.7 * MeV; m2 = 139.6 * MeV
        >>> p = sqrt( Kallen_function( M**2, m1**2, m2**2 ) ) / (2*M)
        >>> print p / GeV   # print the CMS momentum in GeV
        2.61453580221

    References
    ----------
    .. [Kallen] https://en.wikipedia.org/wiki/K%C3%A4ll%C3%A9n_function
    """
    return (x - y - z) ** 2 - 4 * y * z  # faster to evaluate condensed form used


def Armenteros_Podolanski_variables(pplus_3Dvec, pminus_3Dvec):
    """
    Calculate the Armenteros Podolanski [APPaper]_ variables :math:`(\\alpha,q_T)` for a 2-body decay.

    Definition
    ----------

    .. math::

      \\alpha = \\frac{p_L^+ - p_L^-}{p_L^+ + p_L^-}

      q_T = \\frac{| p^- \\times p^{\\mathrm mother}|}{|p^{\\mathrm mother}|}

    where the longitudinal momentum along the direction of flight of the mother particle is

    .. math::

      p_L^\\pm = \\frac{p^\\pm \\cdot p^{\\mathrm mother}}{|p^{\\mathrm mother}|}

    and :math:`q_T` is the transverse momentum of the daughter particles with respect
    to the direction of flight of the mother particle.

    These expressions can be simplified to

    .. math::

      \\alpha = \\frac{|p^+|^2 - |p^-|^2}{|p^+ + p^-|^2}

      q_T = \\frac{| p^+ \\times p^- |}{|p^+ + p^-|}

    Parameters
    -----------
    pplus_3Dvec : Vector3D
        3D-momentum vector of the positively-charged daughter particle.
    pminus_3Dvec : Vector3D
        3D-momentum vector of the negatively-charged daughter particle.

    Returns
    -------
    Tuple :math:`(\\alpha,q_T)`.

    References
    ----------
    .. [APPaper] J. Podolanski and R. Armenteros, III. Analysis of V-events, The London, Edinburgh, and Dublin Philosophical Magazine and Journal of Science 45 (1954) 13, http://dx.doi.org/10.1080/14786440108520416
    """

    mother_mag = (pplus_3Dvec + pminus_3Dvec).mag
    if isequal(mother_mag, 0.0):
        raise ValueError("Total momentum has zero magnitude!")

    # Longitudinal momentum asymmetry, i.e. imbalance
    alpha = (pplus_3Dvec.mag2 - pminus_3Dvec.mag2) / mother_mag**2

    # Transverse momentum of positively-charged particle along the mother particle momentum direction
    qT = (pplus_3Dvec.cross(pminus_3Dvec)).mag / mother_mag

    return (qT, alpha)


def lifetime_to_width(tau):
    """Convert from a particle lifetime to a decay width.

    Parameters
    -----------
    tau : float > 0
        Particle lifetime, typically in picoseconds (any HEP time unit is OK).

    Returns
    -------
    Particle decay width, in the HEP standard energy unit MeV.
    """

    if tau <= 0:
        raise ValueError("Input provided, %s <= 0!".format(tau))

    # Just need to first make sure that the lifetime is in the standard unit ns
    return hbar / float(tau / ns)


def width_to_lifetime(Gamma):
    """Convert from a particle decay width to a lifetime.

    Parameters
    ----------
    Gamma : float > 0
        Particle decay width, typically in MeV (any HEP energy unit is OK).

    Returns
    -------
    Particle lifetime, in the HEP standard time unit ns.
    """

    if Gamma <= 0.0:
        raise ValueError("Input provided, %s <= 0!".format(Gamma))

    # Just need to first make sure that the width is in the standard unit MeV
    return hbar / float(Gamma / MeV)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Submodule for useful exceptions
===============================

.. note:: not meant for user code in general, though possible.
"""

# Definition of handy colours for printing
_default = "\x1b[00m"
_green = "\x1b[01;32m"
_red = "\x1b[01;31m"


class InvalidOperationError(Exception):
    """Exception class for meaningless operations."""

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

    def __str__(self):
        """String representation."""
        return _red + self.message + _default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.

import os

from setuptools import setup

LOCAL_PATH = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(LOCAL_PATH, "requirements.txt")) as requirements_file:
    INSTALL_REQUIRES = requirements_file.read().splitlines()

setup(
    install_requires=INSTALL_REQUIRES,
)



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Vector classes
==============

Two vector classes are available:

* ``Vector3D``     : a 3-dimensional vector.
* ``LorentzVector``: a Lorentz vector, i.e. a 4-dimensional Minkowski space-time vector
                     or a 4-momentum vector.
                     The metric is (-1,-1,-1,+1).
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from skhep.utils.exceptions import *

from math import sqrt, atan2, cos, sin, acos, degrees, log, pi, sinh


# -----------------------------------------------------------------------------
# Vector class in 3D
# -----------------------------------------------------------------------------
class Vector3D(object):
    """
    Vector class in 3 dimensions.

    Constructors:
        __init__(x=0., y=0., z=0.)
        origin()
        frompoint(x, y, z)
        fromvector(avector)
        fromsphericalcoords(r, theta, phi)
        fromcylindricalcoords(rho, phi, z)
        fromiterable(values)
    """

    def __init__(self, x=0.0, y=0.0, z=0.0):
        """Default constructor.

        Examples
        --------
        >>> v1 = Vector3D()
        >>> v1
        <Vector3D (x=0.0,y=0.0,z=0.0)>
        >>> v2 = Vector3D(1,2,3)
        >>> v2
        <Vector3D (x=1,y=2,z=3)>
        """
        self.__values = [x, y, z]

    @classmethod
    def origin(cls):
        """Shortcut constuctor for the origin (x=0.,y=0.,z=0.).
        Equivalent to the default constructor Vector3D().
        """
        return cls(0.0, 0.0, 0.0)

    @classmethod
    def frompoint(cls, x, y, z):
        """Constructor from an explicit space point."""
        return cls(x, y, z)

    @classmethod
    def fromvector(cls, other):
        """Copy constructor."""
        return cls(other.x, other.y, other.z)

    @classmethod
    def fromsphericalcoords(cls, r, theta, phi):
        """Constructor from a space point specified in spherical coordinates.

        Parameters
        ----------
        r     : radius, the radial distance from the origin (r > 0)
        theta : inclination in radians (theta in [0, pi] rad)
        phi   : azimuthal angle in radians (phi in [0, 2pi) rad)
        """
        x = r * sin(theta) * cos(phi)
        y = r * sin(theta) * sin(phi)
        z = r * cos(theta)
        return cls(x, y, z)

    @classmethod
    def fromcylindricalcoords(cls, rho, phi, z):
        """Constructor from a space point specified in cylindrical coordinates.

        Parameters
        ----------
        rho : radial distance from the z-axis (rho > 0)
        phi : azimuthal angle in radians (phi in [-pi, pi) rad)
        z   : height
        """
        x = rho * cos(phi)
        y = rho * sin(phi)
        z = z
        return cls(x, y, z)

    @classmethod
    def fromiterable(cls, values):
        """Constructor from a suitable iterable object.
        Suitable means here that all entries are numbers
        and the length equals 3.
        """
        values = list(values)
        if not len(values) == 3:
            raise ValueError(
                "Input iterable length = {0}! Please check your inputs.".format(
                    len(values)
                )
            )
        for i, v in enumerate(values):
            if not isinstance(v, (int, float)):
                raise ValueError("Component #{0} is not a number!".format(i))
        return cls(values[0], values[1], values[2])

    @property
    def x(self):
        """Return the x, aka first coordinate at position 0."""
        return self.__values[0]

    @x.setter
    def x(self, value):
        """Sets x, aka first coordinate at position 0."""
        self.__values[0] = value

    @property
    def y(self):
        """Return the y, aka second coordinate at position 1."""
        return self.__values[1]

    @y.setter
    def y(self, value):
        """Sets y, aka second coordinate at position 1."""
        self.__values[1] = value

    @property
    def z(self):
        """Return the z, aka third coordinate at position 2."""
        return self.__values[2]

    @z.setter
    def z(self, value):
        """Sets z, aka third coordinate at position 2."""
        self.__values[2] = value

    @property
    def rho(self):
        """Return the cylindrical coordinate rho."""
        return sqrt(self.x**2 + self.y**2)

    @property
    def r(self):
        """Return the spherical coordinate r."""
        return self.mag

    def costheta(self):
        """Return the cosine of the spherical coordinate theta."""
        if self.mag == 0.0:
            return 1.0
        else:
            return self.z / self.mag

    def theta(self, deg=False):
        """Return the spherical coordinate theta.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        theta = acos(self.costheta())
        return theta if not deg else degrees(theta)

    def phi(self, deg=False):
        """Return the spherical or cylindrical coordinate phi.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        phi = atan2(self.y, self.x)
        return phi if not deg else degrees(phi)

    def set(self, x, y, z):
        """Update the vector components all at once."""
        self.__values = [x, y, z]

    def __setitem__(self, i, value):
        """Update/set the ith vector component (commencing at 0, of course)."""
        try:
            self.__values[i] = value
        except IndexError:
            raise IndexError("Vector3D is of length {0} only!".format(len(self)))

    def __getitem__(self, i):
        """Get the ith vector component (commencing at 0, of course)."""
        try:
            return self.__values[i]
        except IndexError:
            raise IndexError("Vector3D is of length {0} only!".format(len(self)))

    def tolist(self):
        """Return the vector as a list."""
        return list(self.__values)

    def __len__(self):
        """Length of the vector, i.e. the number of elements = 3."""
        return len(self.__values)

    @property
    def mag(self):
        """Magnitude, a.k.a. norm, of the vector."""
        return sqrt(self.mag2)

    @property
    def mag2(self):
        """Square of the magnitude, a.k.a. norm, of the vector."""
        return sum(v**2 for v in self.__values)

    def __abs__(self):
        """Get the absolute value for the vector.

        Example
        -------
        >>> v = ...
        >>> a = abs(v)
        """
        return self.mag

    def copy(self):
        """Get a copy of the vector.

        Example
        -------
        >>> v = ...
        >>> v1 = v.copy()
        """
        return Vector3D(self[0], self[1], self[2])

    def unit(self):
        """Return the normalized vector, i.e. the unit vector along the direction of itself."""
        mag = self.mag
        if mag > 0.0 and mag != 1.0:
            return Vector3D.fromiterable([v / mag for v in self.__values])
        else:
            return self

    def __iadd__(self, other):
        """(self)Addition with another vector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 += v2
        """
        if not isinstance(other, Vector3D):
            raise InvalidOperationError(
                "invalid operation '+=' between a 'Vector3D' and a '{0}'".format(
                    other.__class__.__name__
                )
            )
        self.__values[0] += other.__values[0]
        self.__values[1] += other.__values[1]
        self.__values[2] += other.__values[2]
        return self

    def __isub__(self, other):
        """(self)Subtraction with another vector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 -= v2
        """
        if not isinstance(other, Vector3D):
            raise InvalidOperationError(
                "invalid operation '-=' between a 'Vector3D' and a '{0}'".format(
                    other.__class__.__name__
                )
            )
        self.__values[0] -= other.__values[0]
        self.__values[1] -= other.__values[1]
        self.__values[2] -= other.__values[2]
        return self

    def __add__(self, other):
        """Addition with another vector, i.e. self+other."""
        if not isinstance(other, Vector3D):
            return NotImplemented
        v = self.copy()
        v += other
        return v

    def __sub__(self, other):
        """Subtraction with another vector, i.e. self-other."""
        if not isinstance(other, Vector3D):
            return NotImplemented
        v = self.copy()
        v -= other
        return v

    def __imul__(self, other):
        """Scaling of the vector by a number.

        Example
        -------
        >>> v = ...
        >>> v *= 2
        """
        if isinstance(other, (int, float)):
            return Vector3D.fromiterable([v * other for v in self.__values])
        else:
            raise InvalidOperationError(
                "invalid operation '*=' between a 'Vector3D' and a '{0}'".format(
                    other.__class__.__name__
                )
            )

    def __itruediv__(self, number):
        """Scaling of the vector by a number.

        Example
        -------
        >>> v  = ...
        >>> v /= 2
        """
        if not isinstance(number, (int, float)):
            raise InvalidOperationError(
                "invalid operation '/=' between a 'Vector3D' and a '{0}'".format(
                    number.__class__.__name__
                )
            )
        elif 0 == number:
            raise ZeroDivisionError
        self *= 1.0 / number
        return self

    __idiv__ = __itruediv__

    def __mul__(self, other):
        """Multiplication of the vector by either another vector or a number.
        Multiplication of two vectors is equivalent to the dot product, see dot(...).

        Example
        -------
        >>> v2 = v1 * 2
        >>> number = v1 * v3
        """
        if isinstance(other, Vector3D):
            return self.dot(other)
        v = self.copy()
        v *= other
        return v

    def __rmul__(self, other):
        """Right multiplication of the vector by either another vector or a number."""
        return self.__mul__(other)

    def __truediv__(self, number):
        """Division of the vector by a number."""
        v = self.copy()
        v /= number
        return v

    __div__ = __truediv__

    def __eq__(self, other):
        """Equality to another vector, or, equality to zero.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 == v2
        >>> print v1 == 0
        """
        from skhep.math.numeric import isequal

        ## comparsion with scalar zero, very useful  in practice
        if isinstance(other, (float, int)) and isequal(other, 0):
            return isequal(self[0], 0) and isequal(self[1], 0) and isequal(self[2], 0)
        elif not isinstance(other, Vector3D):
            return NotImplemented
        ##
        return (
            isequal(self[0], other[0])
            and isequal(self[1], other[1])
            and isequal(self[2], other[2])
        )

    def __ne__(self, other):
        """Non-equality to another vector.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 != v2
        """
        return not (self == other)

    def __nonzero__(self):
        """Non-zero vector?

        Example
        -------
        >>> vct = ...
        >>> if vct : ...
        """
        return 0 != self.mag2

    __bool__ = __nonzero__

    def __iter__(self):
        """Iterator implementation for the vector components."""
        return self.__values.__iter__()

    def dot(self, other):
        """Dot product with another vector."""
        return sum(v1 * v2 for v1, v2 in zip(self.__values, other.__values))

    def cross(self, other):
        """Cross product with another vector."""
        return Vector3D(
            self[1] * other[2] - self[2] * other[1],
            self[2] * other[0] - self[0] * other[2],
            self[0] * other[1] - self[1] * other[0],
        )

    def rotate(self, angle, *args):
        """Rotate vector by a given angle (in radians) around a given axis."""
        if len(args) == 1 and isinstance(args[0], Vector3D):
            ux, uy, uz = args[0].__values
        elif len(args) == 1 and len(args[0]) == 3:
            ux, uy, uz = args[0]
        elif len(args) == 3:
            ux, uy, uz = args
        else:
            raise TypeError(
                "Input object not a Vector3D nor an iterable with 3 elements."
            )

        for i, u in enumerate((ux, uy, uz)):
            if not isinstance(u, (int, float)):
                raise ValueError("Component #{0} is not a number!".format(i))

        norm = sqrt(ux**2 + uy**2 + uz**2)
        if norm != 1.0:
            ux = ux / norm
            uy = uy / norm
            uz = uz / norm
        c, s = cos(angle), sin(angle)
        c1 = 1.0 - c

        xp = (
            (c + ux**2 * c1) * self.x
            + (ux * uy * c1 - uz * s) * self.y
            + (ux * uz * c1 + uy * s) * self.z
        )
        yp = (
            (ux * uy * c1 + uz * s) * self.x
            + (c + uy**2 * c1) * self.y
            + (uy * uz * c1 - ux * s) * self.z
        )
        zp = (
            (ux * uz * c1 - uy * s) * self.x
            + (uy * uz * c1 + ux * s) * self.y
            + (c + uz**2 * c1) * self.z
        )

        return Vector3D(xp, yp, zp)

    def rotatex(self, angle):
        """Rotate vector by a given angle (in radians) around the x axis."""
        return self.rotate(angle, 1, 0, 0)

    def rotatey(self, angle):
        """Rotate vector by a given angle (in radians) around the y axis."""
        return self.rotate(angle, 0, 1, 0)

    def rotatez(self, angle):
        """Rotate vector by a given angle (in radians) around the z axis."""
        return self.rotate(angle, 0, 0, 1)

    def cosdelta(self, other):
        """Get cos(angle) with respect to another vector"""
        m1 = self.mag2
        if 0 >= m1:
            return 1.0
        m2 = other.mag2
        if 0 >= m2:
            return 1.0

        r = self.dot(other) / sqrt(m1 * m2)
        return max(-1.0, min(1.0, r))

    def angle(self, other, deg=False):
        """
        Angle with respect to another vector.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        cd = self.cosdelta(other)
        return acos(cd) if not deg else degrees(acos(cd))

    def isparallel(self, other):
        """Check if another vector is parallel.
        Two vectors are parallel if they have the same direction but not necessarily the same magnitude.
        """
        from skhep.math.numeric import isequal

        return isequal(self.cosdelta(other), 1)

    def isantiparallel(self, other):
        """Check if another vector is antiparallel.
        Two vectors are antiparallel if they have opposite direction but not necessarily the same magnitude.
        """
        from skhep.math.numeric import isequal

        return isequal(self.cosdelta(other), -1)

    def iscollinear(self, other):
        """Check if another vector is collinear
        Two vectors are collinear if they have parallel or antiparallel
        """
        from skhep.math.numeric import isequal

        return isequal(abs(self.cosdelta(other)), 1)

    def isopposite(self, other):
        """Two vectors are opposite if they have the same magnitude but opposite direction."""
        from skhep.math.numeric import isequal

        added = self + other
        return added == 0

    def isperpendicular(self, other):
        """Check if another vector is perpendicular."""
        from skhep.math.numeric import isequal

        return isequal(self.dot(other), 0, scale=max(self.mag2, other.mag2))

    def __repr__(self):
        """Class representation."""
        return "<Vector3D (x={0},y={1},z={2})>".format(*self.__values)

    def __str__(self):
        """Simple class representation."""
        return str(tuple(self.__values))


# -----------------------------------------------------------------------------
# Lorentz vector class
# -----------------------------------------------------------------------------
class LorentzVector(object):
    """
    Class representing a Lorentz vector,
    either a 4-dimensional Minkowski space-time vector or a 4-momentum vector.
    The 4-vector components can be seen as (x,y,z,t) or (px,py,pz,E).

    Constructors:
        __init__(x=0., y=0., z=0., t=0.)
        from4vector(avector)
        from3vector(vector3d, t)
    """

    def __init__(self, x=0.0, y=0.0, z=0.0, t=0.0):
        """Default constructor.

        Example
        -------
        >>> v1 = LorentzVector()
        >>> v1
        """
        self.__vector3d = Vector3D(x, y, z)
        self.__t = t

    @classmethod
    def from4vector(cls, other):
        """Copy constructor."""
        return cls(other.x, other.y, other.z, other.t)

    @classmethod
    def from3vector(cls, vector3d, t):
        """Constructor from a Vector3D and the time/energy component."""
        return cls(vector3d.x, vector3d.y, vector3d.z, t)

    @classmethod
    def fromiterable(cls, values):
        """Constructor from a suitable iterable object.
        Suitable means here that all entries are numbers
        and the length equals 4.
        """
        values = list(values)
        if not len(values) == 4:
            raise ValueError(
                "Input iterable length = {0}! Please check your inputs.".format(
                    len(values)
                )
            )
        for i, v in enumerate(values):
            if not isinstance(v, (int, float)):
                raise ValueError("Component #{0} is not a number!".format(i))
        return cls(values[0], values[1], values[2], values[3])

    @property
    def x(self):
        """Return the coordinate x, aka first coordinate at position 0."""
        return self.__vector3d.x

    @x.setter
    def x(self, value):
        """Sets x, aka first coordinate at position 0."""
        self.__vector3d.x = value

    @property
    def y(self):
        """Return the coordinate x, aka second coordinate at position 1."""
        return self.__vector3d.y

    @y.setter
    def y(self, value):
        """Sets y, aka second coordinate at position 1."""
        self.__vector3d.y = value

    @property
    def z(self):
        """Return the coordinate z, aka third coordinate at position 2."""
        return self.__vector3d.z

    @z.setter
    def z(self, value):
        """Sets z, aka third coordinate at position 2."""
        self.__vector3d.z = value

    @property
    def vector(self):
        """Return a copy of the vector of spatial components."""
        return self.__vector3d.copy()

    @property
    def t(self):
        """Return the time/energy component, aka coordinate at position 3."""
        return self.__t

    @t.setter
    def t(self, value):
        """Sets t, aka coordinate at position 3."""
        self.__t = value

    def costheta(self):
        """Return the cosinus of the spherical coordinate theta."""
        return self.__vector3d.costheta()

    def theta(self, deg=False):
        """
        Return the spherical coordinate theta.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self.__vector3d.theta(deg)

    def phi(self, deg=False):
        """
        Return the spherical or cylindrical coordinate phi.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self.__vector3d.phi(deg)

    @property
    def px(self):
        """Return the Vector3D coordinate px, aka first momentum coordinate at position 0."""
        return self.__vector3d.x

    @px.setter
    def px(self, value):
        """Sets px, aka first momentum coordinate at position 0."""
        self.__vector3d.x = value

    @property
    def py(self):
        """Return the Vector3D coordinate px, aka second momentum coordinate at position 1."""
        return self.__vector3d.y

    @py.setter
    def py(self, value):
        """Sets py, aka second momentum coordinate at position 1."""
        self.__vector3d.y = value

    @property
    def pz(self):
        """Return the Vector3D coordinate pz, aka third momentum coordinate at position 2."""
        return self.__vector3d.z

    @pz.setter
    def pz(self, value):
        """Sets pz, aka third momentum coordinate at position 2."""
        self.__vector3d.z = value

    @property
    def e(self):
        """Return the energy/time component, aka momentum coordinate at position 3."""
        return self.__t

    @e.setter
    def e(self, value):
        """Sets e, aka momentum coordinate at position 3."""
        self.__t = value

    def set(self, x, y, z, t):
        """Update/set all components at once."""
        self.__vector3d = Vector3D(x, y, z)
        self.__t = t

    def setpxpypzm(self, px, py, pz, m):
        """Set the px,py,pz components and the mass."""
        self.__vector3d.x = px
        self.__vector3d.y = py
        self.__vector3d.z = pz
        if m > 0.0:
            self.__t = sqrt(px**2 + py**2 + pz**2 + m**2)
        else:
            self.__t = sqrt(px**2 + py**2 + pz**2 - m**2)

    def setpxpypze(self, px, py, pz, e):
        """Set the px,py,pz components and the energy."""
        self.set(px, py, pz, e)

    def setptetaphim(self, pt, eta, phi, m):
        """Set the transverse momentum, the pseudorapidity, the angle phi and the mass."""
        px, py, pz = pt * cos(phi), pt * sin(phi), pt * sinh(eta)
        self.setpxpypzm(px, py, pz, m)

    def setptetaphie(self, pt, eta, phi, e):
        """Set the transverse momentum, the pseudorapidity, the angle phi and the energy."""
        px, py, pz = pt * cos(phi), pt * sin(phi), pt * sinh(eta)
        self.setpxpypze(px, py, pz, e)

    def tolist(self):
        """Return the LorentzVector as a list."""
        return list(self.__vector3d) + [self.__t]

    def __setitem__(self, i, value):
        """Update/set the ith vector component (commencing at 0, of course)."""
        try:
            if i == 3:
                self.__t = value
            else:
                self.__vector3d[i] = value
        except IndexError:
            raise IndexError("LorentzVector is of length {0} only!".format(len(self)))

    def __getitem__(self, i):
        """Get the ith vector component (commencing at 0, of course)."""
        try:
            return self.tolist()[i]
        except IndexError:
            raise IndexError("LorentzVector is of length {0} only!".format(len(self)))

    def __len__(self):
        """Length of the LorentzVector, i.e. the number of elements = 4."""
        return len(self.tolist())

    @property
    def p(self):
        """Return the momentum, aka norm of the momentum vector."""
        return self.__vector3d.mag

    @property
    def pt(self):
        """Return the transverse momentum, aka transverse component of the momentum vector."""
        return self.perp

    @property
    def et(self):
        """Return the transverse energy."""
        return self.e * (self.pt / self.p)

    @property
    def m(self):
        """Return the invariant mass."""
        return self.mag

    @property
    def m2(self):
        """Return the square of the invariant mass."""
        return self.mag2

    @property
    def mass(self):
        """Return the invariant mass."""
        return self.mag

    @property
    def mass2(self):
        """Return the square of the invariant mass."""
        return self.mag2

    @property
    def mt(self):
        """Return the transverse mass."""
        return self.transversemass

    @property
    def mt2(self):
        """Return the square of the transverse mass."""
        return self.transversemass2

    @property
    def transversemass(self):
        """Return the transverse mass."""
        mt2 = self.transversemass2
        return sqrt(mt2) if mt2 >= 0.0 else -sqrt(-mt2)

    @property
    def transversemass2(self):
        """Return the square of the transverse mass."""
        return self.e**2 - self.pz**2

    @property
    def beta(self):
        """Return :math:`\\beta = v/c`."""
        return self.p / self.e

    @property
    def gamma(self):
        """Return :math:`\\gamma = 1/\\sqrt{1-\\beta^2}`."""
        if self.beta < 1:
            return 1.0 / sqrt(1.0 - self.beta**2)
        else:
            return 10e10

    @property
    def eta(self):
        """Return the pseudorapidity."""
        if abs(self.costheta()) < 1.0:
            return -0.5 * log((1.0 - self.costheta()) / (1.0 + self.costheta()))
        else:
            return 10e10 if self.z > 0 else -10e10

    @property
    def boostvector(self):
        """Return the spatial component divided by the time component."""
        return Vector3D(self.x / self.t, self.y / self.t, self.z / self.t)

    @property
    def pseudorapidity(self):
        """ "Return the pseudorapidity. Alternative to eta() method."""
        return self.eta

    @property
    def rapidity(self):
        """Return the rapidity."""
        return 0.5 * log((self.e + self.pz) / (self.e - self.pz))

    @property
    def mag(self):
        """Magnitude, a.k.a. norm, of the Lorentz vector."""
        mag2 = self.mag2
        return sqrt(mag2) if mag2 >= 0.0 else -sqrt(-mag2)

    @property
    def mag2(self):
        """Square of the magnitude, a.k.a. norm, of the Lorentz vector."""
        return self.t**2 - self.__vector3d.mag2

    @property
    def perp2(self):
        """Square of the transverse component, in the (x,y) plane, of the spatial components."""
        return self.x**2 + self.y**2

    @property
    def perp(self):
        """Transverse component of the spatial components."""
        return sqrt(self.perp2)

    def copy(self):
        """Get a copy of the LorentzVector.

        Example
        -------
        >>> v = ...
        >>> v1 = v.copy()
        """
        return LorentzVector(self[0], self[1], self[2], self[3])

    def __iadd__(self, other):
        """(self)Addition with another LorentzVector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 += v2
        """
        if not isinstance(other, LorentzVector):
            raise InvalidOperationError(
                "invalid operation '+=' between a 'LorentzVector' and a '{0}'".format(
                    other.__class__.__name__
                )
            )
        self.__vector3d += other.__vector3d
        self.__t += other.__t
        return self

    def __isub__(self, other):
        """(self)Subtraction with another LorentzVector, i.e. self+other.

        Example
        -------
        >>> v1  = ...
        >>> v2  = ...
        >>> v1 -= v2
        """
        if not isinstance(other, LorentzVector):
            raise InvalidOperationError(
                "invalid operation '-=' between a 'LorentzVector' and a '{0}'".format(
                    other.__class__.__name__
                )
            )
        self.__vector3d -= other.__vector3d
        self.__t -= other.__t
        return self

    def __add__(self, other):
        """Addition with another LorentzVector, i.e. self+other."""
        v = self.copy()
        v += other
        return v

    def __sub__(self, other):
        """Subtraction with another LorentzVector, i.e. self-other."""
        v = self.copy()
        v -= other
        return v

    def __imul__(self, other):
        """Scaling of the Lorentz vector with a number.

        Example
        -------
        >>> v = ...
        >>> v *= 2
        """
        if isinstance(other, (int, float)):
            return LorentzVector.fromiterable([v * other for v in self.tolist()])
        else:
            raise InvalidOperationError(
                "invalid operation '*=' between a 'LorentzVector' and a '{0}'".format(
                    other.__class__.__name__
                )
            )

    def __itruediv__(self, number):
        """Scaling of the Lorentz vector with a number.

        Example
        -------
        >>> v  = ...
        >>> v /= 2
        """
        if not isinstance(number, (int, float)):
            raise InvalidOperationError(
                "invalid operation '/=' between a 'LorentzVector' and a '{0}'".format(
                    number.__class__.__name__
                )
            )
        elif 0 == number:
            raise ZeroDivisionError
        self *= 1.0 / number
        return self

    __idiv__ = __itruediv__

    def __mul__(self, other):
        """Multiplication of the LorentzVector by either another LorentzVector or a number.
        Multiplication of two LorentzVector is equivalent to the dot product, see dot(...).

        Example
        -------
        >>> v2 = v1 * 2
        >>> number = v1 * v3
        """
        if isinstance(other, LorentzVector):
            return self.dot(other)
        v = self.copy()
        v *= other
        return v

    def __rmul__(self, other):
        """Right multiplication of the LorentzVector by either another LorentzVector or a number."""
        return self.__mul__(other)

    def __truediv__(self, number):
        """Division of the LorentzVector by a number."""
        v = self.copy()
        v /= number
        return v

    __div__ = __truediv__

    def __eq__(self, other):
        """Equality to another LorentzVector, or, equality to zero.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 == v2
        >>> print v1 == 0
        """
        from skhep.math.numeric import isequal

        ## comparsion with scalar zero, very useful  in practice
        if isinstance(other, (float, int)) and isequal(other, 0):
            return (
                isequal(self[0], 0)
                and isequal(self[1], 0)
                and isequal(self[2], 0)
                and isequal(self[3], 0)
            )
        elif not isinstance(other, LorentzVector):
            return NotImplemented
        ##
        return (
            isequal(self[0], other[0])
            and isequal(self[1], other[1])
            and isequal(self[2], other[2])
            and isequal(self[3], other[3])
        )

    def __ne__(self, other):
        """Non-equality to another Lorentz vector.

        Example
        -------
        >>> v1 = ...
        >>> v2 = ...
        >>> print v1 != v2
        """
        return not (self == other)

    def __iter__(self):
        """Iterator implementation for the Lorentz vector components."""
        return self.tolist().__iter__()

    def boost(self, *args):
        """Apply a Lorentz boost on the Lorentz vector."""
        if len(args) == 1 and isinstance(args[0], Vector3D):
            bx, by, bz = args[0].x, args[0].y, args[0].z
        elif len(args) == 1 and len(args[0]) == 3:
            bx, by, bz = args[0]
        elif len(args) == 3:
            bx, by, bz = args
        else:
            raise TypeError(
                "Input object not a Vector3D nor an iterable with 3 elements."
            )

        for i, b in enumerate((bx, by, bz)):
            if not isinstance(b, (int, float)):
                raise ValueError("Component #{0} is not a number!".format(i))

        b2 = bx**2 + by**2 + bz**2
        gamma = 1.0 / sqrt(1.0 - b2)
        bp = bx * self.x + by * self.y + bz * self.z
        if b2 > 0.0:
            gamma2 = (gamma - 1.0) / b2
        else:
            gamma2 = 0.0

        xp = self.x + gamma2 * bp * bx - gamma * bx * self.t
        yp = self.y + gamma2 * bp * by - gamma * by * self.t
        zp = self.z + gamma2 * bp * bz - gamma * bz * self.t
        tp = gamma * (self.t - bp)

        return LorentzVector(xp, yp, zp, tp)

    def rotate(self, angle, *args):
        """Rotate vector by a given angle (in radians) around a given axis."""

        v3p = self.__vector3d.rotate(angle, *args)

        return LorentzVector().from3vector(v3p, self.t)

    def rotatex(self, angle):
        """Rotate vector by a given angle (in radians) around the x axis."""
        return self.rotate(angle, 1, 0, 0)

    def rotatey(self, angle):
        """Rotate vector by a given angle (in radians) around the y axis."""
        return self.rotate(angle, 0, 1, 0)

    def rotatez(self, angle):
        """Rotate vector by a given angle (in radians) around the z axis."""
        return self.rotate(angle, 0, 0, 1)

    def dot(self, other):
        """Dot product with another Lorentz vector."""
        return self.t * other.t - self.__vector3d * other.__vector3d

    def deltaeta(self, other):
        """Return the pseudorapidity difference, :math:`\\Delta \\eta`, with another Lorentz vector."""
        return self.eta - other.eta

    def deltaphi(self, other):
        """Return the phi angle difference, :math:`\\Delta \\phi`, with another Lorentz vector."""
        dphi = self.phi() - other.phi()
        while dphi > pi:
            dphi -= 2 * pi
        while dphi < -pi:
            dphi += 2 * pi
        return dphi

    def deltar(self, other):
        """Return :math:`\\Delta R` the distance in (eta,phi) space with another Lorentz vector, defined as:
        :math:`\\Delta R = \\sqrt{(\\Delta \\eta)^2 + (\\Delta \\phi)^2}`
        """
        return sqrt(self.deltaeta(other) ** 2 + self.deltaphi(other) ** 2)

    def isspacelike(self):
        """Check if Lorentz Vector is space-like."""
        from skhep.math.numeric import isequal

        return self.mag2 < 0.0 and not isequal(self.mag2, 0.0)

    def istimelike(self):
        """Check if Lorentz Vector is time-like."""
        from skhep.math.numeric import isequal

        return self.mag2 > 0.0 and not isequal(self.mag2, 0.0)

    def islightlike(self):
        """Check if Lorentz Vector is light-like."""
        from skhep.math.numeric import isequal

        return isequal(self.mag2, 0.0)

    def __repr__(self):
        """Class representation."""
        return "<LorentzVector (x={0},y={1},z={2},t={3})>".format(*self.tolist())

    def __str__(self):
        """Simple class representation."""
        return str(tuple(self.tolist()))
